use hex_color::HexColor;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use uuid::Uuid;

use super::{Event, EventId};

pub type CalendarId = String;

pub const CALENDAR_PROP_NAME: &str = "name";
pub const CALENDAR_PROP_ENABLED: &str = "enabled";
pub const CALENDAR_PROP_DEFAULT: &str = "default";

#[derive(Debug, Clone, Deserialize, Serialize, PartialEq, Eq)]
#[gobject(id, name)]
pub struct Calendar {
    pub id: String,
    pub name: String,
    pub events: HashMap<EventId, Event>,
    pub enabled: bool,
    pub default: bool,
    pub color: String,
    pub is_remote: bool,
}

pub const DEFAULT_CALENDAR_COLOR: &str = "#0E403560";

impl Default for Calendar {
    fn default() -> Self {
        Self {
            id: Uuid::new_v4().to_string(),
            name: "New Calendar".into(),
            events: Default::default(),
            enabled: true,
            default: false,
            color: DEFAULT_CALENDAR_COLOR.into(),
            is_remote: false,
        }
    }
}

pub fn adjust_color(input_color: &str) -> String {
    if let Ok(color) = HexColor::parse(input_color) {
        // 96 dec = 60 hex
        if color.a > 96 {
            return format!("{}60", &input_color[..7]);
        }
    }
    input_color.to_string()
}

pub fn get_text_color(bg_color: &str) -> String {
    if let Ok(color) = HexColor::parse(bg_color) {
        return get_contrast_color(
            color.r as f64 / 255.0,
            color.g as f64 / 255.0,
            color.b as f64 / 255.0,
            color.a as f64 / 255.0,
        );
    }
    "black".into()
}

fn get_contrast_color(r: f64, g: f64, b: f64, a: f64) -> String {
    let alpha = 1.0 - a;
    let brightness = (r * 0.299 + g * 0.587 + b * 0.114 + (alpha)) * 255.0;
    if brightness > 186.0 {
        "black".into()
    } else {
        "white".into()
    }
}

#[cfg(test)]
mod tests {
    use crate::domain::get_text_color;

    use super::adjust_color;

    #[test]
    fn test_color_adjustment() {
        assert_eq!(adjust_color("#00ff33"), "#00ff3360");
        assert_eq!(adjust_color("#ff00f090"), "#ff00f060");
    }

    #[test]
    fn test_get_text_color() {
        assert_eq!(get_text_color("#ffff33"), "black");
        assert_eq!(get_text_color("#0000ff"), "white");
    }
}
