use crate::store::{persistence::get_app_dir, store, State};
use gdk4::subclass::prelude::ObjectSubclassIsExt;
use glib::PRIORITY_DEFAULT_IDLE;
use gtk::{prelude::*, TemplateChild};
use gtk_rust_app::widgets::Page;
use percent_encoding::{utf8_percent_encode, AsciiSet, CONTROLS};

pub const FEEDBACK_PAGE_NAME: &str = "feedback";

const MAILTO: &str = "mailto:contact-project+loers-karlender-32697518-issue-@incoming.gitlab.com";

const MAIL_TEMPLATE: &str = r#"
## Summary

(Summarize the bug encountered concisely)

## Steps to reproduce

(How one can reproduce the issue - this is very important)

## What is the current bug behavior?

(What actually happens)

## What is the expected correct behavior?

(What you should see instead)

## Relevant logs and/or screenshots


<details>
  <summary>My latest logs</summary>

  ```
{logs}
```
</details>
"#;

#[widget(@store extends gtk::Box)]
#[template(file = "feedback.ui")]
pub struct FeedbackPage {
    #[template_child]
    pub buttons_box: TemplateChild<gtk::Box>,

    #[template_child]
    pub submit_issue_button: TemplateChild<gtk::Button>,
    #[template_child]
    pub spinner: TemplateChild<gtk::Spinner>,

    #[template_child]
    pub logs_button: TemplateChild<gtk::ToggleButton>,

    #[template_child]
    pub logs_textview: TemplateChild<gtk::TextView>,

    #[template_child]
    pub mailto_button: TemplateChild<gtk::LinkButton>,

    #[selector(state.ui.mobile)]
    select_mobile: (),
}

impl FeedbackPage {
    pub fn new() -> Self {
        let o: Self = glib::Object::new();
        o
    }
    pub fn constructed(&self) {
        let _self = self;
        // let mailto_button = self.mailto_button();
        // let spinner = self.spinner();
        self.mailto_button().set_uri(MAILTO);

        let (send, recv) = glib::MainContext::channel(PRIORITY_DEFAULT_IDLE);

        recv.attach(
            None,
            glib::clone!(@weak _self => @default-return glib::Continue(false), move |_| {
                //
                _self.submit_issue_button().set_sensitive(true);
                _self.spinner().set_visible(false);
                glib::Continue(true)
            }),
        );

        self.submit_issue_button()
            .connect_clicked(glib::clone!(@weak _self => move |_| {
                _self.spinner().set_visible(true);
                _self.submit_issue_button().set_sensitive(false);
                const FRAGMENT: &AsciiSet = &CONTROLS.add(b' ').add(b'"').add(b'<').add(b'>').add(b'`');
                let url = format!("{}?body={}", MAILTO, MAIL_TEMPLATE.replace("{logs}", &get_log_messages()));
                let url = utf8_percent_encode(&url, FRAGMENT);
                _self.mailto_button().set_uri(
                    &url.to_string()
                );
                _self.mailto_button().emit_clicked();
                _self.mailto_button().set_uri(MAILTO);
                let send = send.clone();
                std::thread::spawn(move || {
                    std::thread::sleep(std::time::Duration::from_secs(5));
                    if let Err(e) = send.send(()) {
                        error!("Internal error: {}", e)
                    }
                });
            }));

        let logs_textview = self.logs_textview();
        logs_textview.set_visible(false);

        self.logs_button()
            .connect_clicked(glib::clone!(@weak logs_textview => move |b| {
                let buffer = logs_textview.buffer();
                buffer.set_text(&get_log_messages());
                logs_textview.emit_select_all(true);
                logs_textview.set_visible(b.is_active());
            }));
    }

    fn select_mobile(&self, state: &State) {
        if state.ui.mobile {
            self.buttons_box()
                .set_orientation(gtk::Orientation::Vertical);
            self.buttons_box().set_halign(gtk::Align::Fill);
        } else {
            self.buttons_box()
                .set_orientation(gtk::Orientation::Horizontal);
            self.buttons_box().set_halign(gtk::Align::Center);
        }
    }
}

impl Page for FeedbackPage {
    fn name(&self) -> &'static str {
        FEEDBACK_PAGE_NAME
    }

    fn title_and_icon(&self) -> Option<(String, String)> {
        None
    }
}

impl Default for FeedbackPage {
    fn default() -> Self {
        Self::new()
    }
}

fn get_log_messages() -> String {
    if let Ok(logs) = std::fs::read_to_string(get_app_dir().join("karlender.log")) {
        logs
    } else {
        "No logs found".into()
    }
}
