use crate::domain::*;
use crate::store::selection::Selection;
use crate::store::{
    gstore::{dispatch, select},
    store, State,
};
use crate::widgets::{WeekEventTag, WeekHeaderCell, WeekHourHeaderCell};
use gdk4::subclass::prelude::ObjectSubclassIsExt;
use gtk::ffi::GTK_STYLE_PROVIDER_PRIORITY_APPLICATION;
use gtk::{prelude::*, CssProvider};
use std::cell::Cell;

use super::{get_month_by_num, WeekCell};

const CELL_HEIGHT: i32 = 20;
const MAX_COLS: i32 = 2;

#[widget(extends gtk::Box)]
#[template(file = "week_grid.ui")]
pub struct WeekGrid {
    #[property_i64]
    delta: Cell<i64>,
    #[property_string]
    title: Cell<String>,

    pub timezone: Cell<Option<KarlenderTimezone>>,

    #[template_child]
    pub title_label: TemplateChild<gtk::Label>,
    #[template_child]
    pub days_header_grid: TemplateChild<gtk::Grid>,
    #[template_child]
    pub hour_header_grid: TemplateChild<gtk::Grid>,
    #[template_child]
    pub grid: TemplateChild<gtk::Grid>,
    #[template_child]
    pub hour_indicator: TemplateChild<gtk::Box>,

    pub selection: Cell<Option<KarlenderDateTime>>,
}

impl WeekGrid {
    pub fn new(delta: i64) -> Self {
        let o: WeekGrid = glib::Object::new();
        o.set_properties(&[
            //
            ("delta", &delta),
            ("title", &""),
        ]);
        o
    }

    pub fn title(&self) -> String {
        self.property("title")
    }

    pub fn constructed(&self) {
        self.connect_realize(|_self| {
            _self.setup_grid();
            select!(
                |state| state.ui =>
                glib::clone!(@weak _self => move |state| {
                    _self.on_ui_state_update(state);
                })
            );
            select!(
                |state|
                    state.calendar_slice,
                    state.settings,
                    state.ui.mobile,
                => glib::clone!(@weak _self => move |state| {
                    _self.on_calendars_update(state);
                })
            );
        });
    }

    fn on_ui_state_update(&self, state: &State) {
        {
            if self.property::<i64>("delta") == 0 {
                self.hour_indicator().set_visible(true)
            } else {
                self.hour_indicator().set_visible(false)
            }
            let (h, m) = state.ui.time;
            self.hour_indicator()
                .set_margin_top(height_from_time(h as i32 * 60 + m as i32));
        }
        if state.ui.mobile {
            self.title_label().hide();
        } else {
            self.title_label().show();
        }
    }

    fn on_calendars_update(&self, state: &State) {
        let delta: i64 = self.property("delta");
        let today = KarlenderTimezone::local().today();
        let delta_today = if delta >= 0 {
            today + KarlenderDateTimeDuration::days(delta * 7)
        } else {
            today - KarlenderDateTimeDuration::days(delta.abs() * 7)
        };
        let tz = state
            .settings
            .timezone
            .as_ref()
            .map(|s| KarlenderTimezone::from_str(s));
        self.imp().timezone.set(tz.clone());

        let num_from_monday = delta_today.weekday().number_from_monday() as i64;
        let day_0 = delta_today.clone() - KarlenderDateTimeDuration::days(num_from_monday);
        let day_1 = delta_today.clone() - KarlenderDateTimeDuration::days(num_from_monday - 1);
        let day_2 = delta_today.clone() - KarlenderDateTimeDuration::days(num_from_monday - 2);
        let day_3 = delta_today.clone() - KarlenderDateTimeDuration::days(num_from_monday - 3);
        let day_4 = delta_today.clone() - KarlenderDateTimeDuration::days(num_from_monday - 4);
        let day_5 = delta_today.clone() - KarlenderDateTimeDuration::days(num_from_monday - 5);
        let day_6 = delta_today - KarlenderDateTimeDuration::days(num_from_monday - 6);

        let mut day_0_events = Vec::new();
        let mut day_1_events = Vec::new();
        let mut day_2_events = Vec::new();
        let mut day_3_events = Vec::new();
        let mut day_4_events = Vec::new();
        let mut day_5_events = Vec::new();
        let mut day_6_events = Vec::new();

        let mut child = self.grid().first_child();
        while let Some(c) = child {
            child = c.next_sibling();
            if let Ok(event) = c.downcast::<WeekEventTag>() {
                self.grid().remove(&event);
            }
        }

        for calendar in state.calendar_slice.calendars.values() {
            for event in calendar.events.values() {
                if event.full_day {
                    let day = event.start.date();
                    if day == day_0
                        || day == day_1
                        || day == day_2
                        || day == day_3
                        || day == day_4
                        || day == day_5
                        || day == day_6
                    {
                        let col = event.start.weekday().number_from_monday() as i32 * MAX_COLS;
                        let full_day_event = gtk::Box::builder()
                            .css_classes(vec!["event-tag".to_string(), "osd".into()])
                            .build();
                        let full_day_event_label = gtk::Label::builder()
                            .label(&event.name)
                            .hexpand(true)
                            .ellipsize(gdk4::pango::EllipsizeMode::End)
                            .build();
                        full_day_event_label.add_css_class("px-1");

                        let css = CssProvider::new();

                        css.load_from_data(&format!(
                            "* {{ background-color: {}; color: {};}}",
                            &calendar.color,
                            get_text_color(&calendar.color)
                        ));

                        full_day_event_label
                            .style_context()
                            .add_provider(&css, GTK_STYLE_PROVIDER_PRIORITY_APPLICATION as u32);

                        full_day_event.append(&full_day_event_label);

                        self.grid().attach(&full_day_event, col, 0, MAX_COLS, 1);
                    }
                    continue;
                }

                self.layout_event_tag(0, &day_0, calendar, event, &tz, &mut day_0_events);
                self.layout_event_tag(1, &day_1, calendar, event, &tz, &mut day_1_events);
                self.layout_event_tag(2, &day_2, calendar, event, &tz, &mut day_2_events);
                self.layout_event_tag(3, &day_3, calendar, event, &tz, &mut day_3_events);
                self.layout_event_tag(4, &day_4, calendar, event, &tz, &mut day_4_events);
                self.layout_event_tag(5, &day_5, calendar, event, &tz, &mut day_5_events);
                self.layout_event_tag(6, &day_6, calendar, event, &tz, &mut day_6_events);
            }
        }

        self.layout_events_in_day(MAX_COLS, &mut day_0_events);
        self.layout_events_in_day(MAX_COLS, &mut day_1_events);
        self.layout_events_in_day(MAX_COLS, &mut day_2_events);
        self.layout_events_in_day(MAX_COLS, &mut day_3_events);
        self.layout_events_in_day(MAX_COLS, &mut day_4_events);
        self.layout_events_in_day(MAX_COLS, &mut day_5_events);
        self.layout_events_in_day(MAX_COLS, &mut day_6_events);

        let sort_events = |a: &LayoutedEventTag, b: &LayoutedEventTag| -> std::cmp::Ordering {
            match (b.end - b.start).cmp(&(a.end - a.start)) {
                std::cmp::Ordering::Equal => a.start.cmp(&b.start),
                o => o,
            }
        };

        day_0_events.sort_by(sort_events);
        day_1_events.sort_by(sort_events);
        day_2_events.sort_by(sort_events);
        day_3_events.sort_by(sort_events);
        day_4_events.sort_by(sort_events);
        day_5_events.sort_by(sort_events);
        day_6_events.sort_by(sort_events);

        for e in day_0_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
        for e in day_1_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
        for e in day_2_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
        for e in day_3_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
        for e in day_4_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
        for e in day_5_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
        for e in day_6_events {
            self.grid()
                .attach(&e.event_tag, e.column, e.start, e.width, e.end - e.start);
        }
    }

    fn layout_event_tag(
        &self,
        weekday: i32,
        day: &KarlenderDate,
        calendar: &Calendar,
        event: &Event,
        tz: &Option<KarlenderTimezone>,
        events: &mut Vec<LayoutedEventTag>,
    ) {
        if &event.start.date() == day {
            let h = tz.as_ref().unwrap().from_utc(&event.start).hour();
            let m = tz.as_ref().unwrap().from_utc(&event.start).minute();
            let row = h as i32 * 4 + m as i32 / CELL_HEIGHT;
            let duration = event.end.signed_duration_since(event.start.clone());
            let mut height = duration.num_minutes() as i32 / CELL_HEIGHT;
            if height == 0 {
                height = 1;
            }
            let event_tag = WeekEventTag::new(event.name.clone(), 1, &calendar.color, false);
            event_tag.add_css_class("week-event");
            event_tag.set_cursor(gdk4::Cursor::from_name("pointer", None).as_ref());

            let calendar_id = calendar.id.clone();
            let event_id = event.id.clone();
            event_tag.connect_click(move |_| {
                dispatch!(
                    crate::store::EDIT_EVENT,
                    (calendar_id.clone(), event_id.clone())
                );
                dispatch!(crate::store::NAVIGATE, crate::pages::EDITOR_PAGE_NAME);
            });
            let le = LayoutedEventTag {
                column: weekday * MAX_COLS,
                width: MAX_COLS,
                start: 1 + row, // the events grid starts with 1 due to the 0th row beeing for whole-day events
                end: 1 + row + height,
                event_tag,
            };
            if le.end - le.start > 2 {
                le.event_tag.set_long(true);
            }
            events.push(le);
        }
    }

    fn layout_events_in_day(&self, max_cols: i32, events: &mut [LayoutedEventTag]) {
        let iter_events: Vec<(i32, i32, i32, i32)> = events
            .iter()
            .map(|e| (e.column, e.width, e.start, e.end))
            .collect();

        for a in 0..iter_events.len() {
            let (a_col, a_width, a_start, a_end) = &iter_events[a];
            for b in a..iter_events.len() {
                let (b_col, b_width, b_start, b_end) = &iter_events[b];

                if a == b {
                    continue;
                }

                let overlaps_in_y_axis =
                    a_start >= b_start && a_start < b_end || a_end > b_start && a_end <= b_end;

                let overlaps_in_x_axis = (a_col >= b_col && *a_col < b_col + b_width)
                    || (a_col + a_width > *b_col && a_col + a_width <= b_col + b_width);

                if overlaps_in_y_axis
                    && overlaps_in_x_axis
                    && a_start == b_start
                    && *a_width == max_cols
                    && *b_width == max_cols
                {
                    events[a].width /= 2;
                    events[b].width /= 2;
                    events[b].column = max_cols / 2;
                }
            }
        }
    }

    pub fn setup_grid(&self) {
        let _self = self;
        let grid = &self.grid();
        let hour_header_grid = &self.hour_header_grid();
        let delta: i64 = self.property("delta");
        let today = KarlenderTimezone::local().today();
        let month_len = get_month_len(&today) as i64;
        let former_month = if today.month0() == 0 {
            KarlenderDate::from_ymd(today.year() - 1, 12, 1)
        } else {
            KarlenderDate::from_ymd(today.year(), today.month() - 1, 1)
        };
        let former_month_len = get_month_len(&former_month) as i64;
        let delta_today = if delta >= 0 {
            today + KarlenderDateTimeDuration::days(delta * 7)
        } else {
            today - KarlenderDateTimeDuration::days(delta.abs() * 7)
        };
        let delta_day = delta_today.day() as i64;
        let delta_weekday = delta_today.weekday().number_from_monday() as i64;

        let first_day_of_week = delta_today.clone()
            - KarlenderDateTimeDuration::days(delta_today.weekday().number_from_monday() as i64);
        let last_day_of_week = delta_today.clone()
            + KarlenderDateTimeDuration::days(
                6 - delta_today.weekday().number_from_monday() as i64,
            );

        match first_day_of_week.month().cmp(&last_day_of_week.month()) {
            std::cmp::Ordering::Less => {
                self.set_property(
                    "title",
                    format!(
                        "{} {}/{}",
                        first_day_of_week.year(),
                        get_month_by_num(first_day_of_week.month()),
                        get_month_by_num(last_day_of_week.month()),
                    ),
                );
            }
            std::cmp::Ordering::Greater => {
                self.set_property(
                    "title",
                    format!(
                        "{} {}/{}",
                        first_day_of_week.year(),
                        get_month_by_num(first_day_of_week.month()),
                        get_month_by_num(last_day_of_week.month()),
                    ),
                );
            }
            std::cmp::Ordering::Equal => {
                self.set_property(
                    "title",
                    format!(
                        "{} {}",
                        delta_today.year(),
                        get_month_by_num(delta_today.month()),
                    ),
                );
            }
        }

        let corner = gtk::Box::builder().width_request(24).build();
        self.days_header_grid().attach(&corner, 0, 0, 1, 1);
        for i in 0..7 {
            let mut day = delta_day - delta_weekday + i;
            if day <= 0 {
                day += former_month_len - 1;
            }
            if day >= month_len {
                day -= month_len - 1;
            }
            let header = WeekHeaderCell::new(i as u64, day as u64);
            if delta == 0 && day == delta_day {
                header.add_css_class("cell-current");
            }
            self.days_header_grid()
                .attach(&header, 1 + i as i32, 0, 1, 1);
        }

        let tz = self.imp().timezone.take();
        self.imp().timezone.set(tz.clone());

        let full_day_cell = WeekCell::new(last_day_of_week.and_hms(0, 1, 0, tz.as_ref()));
        full_day_cell.set_height_request(20);
        full_day_cell.set_vexpand(false);
        hour_header_grid.attach(&full_day_cell, 0, 0, 1, 1);
        for j in 1..24 {
            let header = WeekHourHeaderCell::new(j as u64);
            hour_header_grid.attach(&header, 0, j, 1, 1);
        }
        for i in 0i32..7 {
            let day = first_day_of_week.clone() + KarlenderDateTimeDuration::days(i as i64);
            let full_day_cell = WeekCell::new(day.and_hms(0, 1, 0, tz.as_ref()));
            full_day_cell.set_height_request(20);
            grid.attach(&full_day_cell, MAX_COLS * i, 0, MAX_COLS, 1);
            for j in 0..96i32 {
                let hour = (j / 4) as u8;
                let minutes = j.rem_euclid(4) as u8 * 15;
                let datetime = day.and_hms(hour, minutes, 0, tz.as_ref());
                let cell = WeekCell::new(datetime);
                cell.connect_clicked(glib::clone!(@weak _self => move |cell, _| {
                    let mut child = _self.grid().first_child();
                    while let Some(c) = child {
                        child = c.next_sibling();
                        if let Ok(cell) = c.downcast::<WeekCell>() {
                            cell.remove_css_class("week-cell-selected");
                        }
                    }
                    cell.add_css_class("week-cell-selected");
                    let datetime = cell.datetime();
                    let selection = _self.imp().selection.take();
                    match selection {
                        None => {
                            _self.imp().selection.set(Some(datetime));
                        },
                        Some(s) => {
                            if s == datetime {
                                let timezone = _self.imp().timezone.take();
                                _self.imp().timezone.set(timezone.clone());
                                if let Some(tz) = timezone.as_ref() {
                                    let sel = Selection::new(
                                        tz.from_utc(&s),
                                        tz.from_utc(&(datetime + KarlenderDateTimeDuration::minutes(15))),
                                        false,
                                        true,
                                        1);
                                    dispatch!(crate::store::SELECT, sel.to_variant());
                                    dispatch!(crate::store::NEW_EVENT)
                                }
                            }
                        }
                    }
                }));
                grid.attach(&cell, MAX_COLS * i, 1 + j, MAX_COLS, 1);
            }
        }
    }
}

fn height_from_time(minutes: i32) -> i32 {
    minutes / 15 * CELL_HEIGHT
}

#[derive(Debug)]
struct LayoutedEventTag {
    column: i32,
    width: i32,
    start: i32,
    end: i32,
    event_tag: WeekEventTag,
}
